//===- LuaBindingsConsumer.cpp ---------------------------------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// LuaBindingsConsumer.cpp - Declaration of LuaBindingsConsumer.
//
//===----------------------------------------------------------------------===//

#include "LuaBindings.h"

namespace LuaBindings {

using namespace clang;

// FunctionOverloads members.

FunctionOverloads::FunctionOverloads(
      const clang::FunctionDecl *function,
      std::string functionName,
      bool isMethod)
    : FunctionName(functionName), IsMethod(isMethod) {
  Overloads.push_back(function);
}

FunctionOverloads::~FunctionOverloads() {
}

// DeclarationState members.

DeclarationState::DeclarationState(
      const clang::NamedDecl *decl,
      std::string &typeName,
      std::string &qualifiedTypeName,
      std::string &bindTypeName) :
    Declaration(decl),
    TypeName(typeName),
    QualifiedTypeName(qualifiedTypeName),
    BindTypeName(bindTypeName),
    BindMode(BindAll) {
}

DeclarationState::DeclarationState(std::string bindTypeName) :
    Declaration(NULL),
    BindTypeName(bindTypeName),
    BindMode(BindAll) {
}

DeclarationState::~DeclarationState() {
  int index;
  for (index = Functions.size() - 1; index >= 0; index--)
    delete Functions[index];
  Functions.resize(0);
}

  // Find a function overload entry.
FunctionOverloads *DeclarationState::FindFunction(const std::string &name) {
  std::vector<FunctionOverloads *>::iterator iter;
  std::vector<FunctionOverloads *>::iterator end = Functions.end();
  for (iter = Functions.begin(); iter != end; ++iter) {
    if ((*iter)->FunctionName == name)
      return *iter;
  }
  return NULL;
}

// LuaBindingsConsumer members.

  // Constructor.
LuaBindingsConsumer::LuaBindingsConsumer(
        clang::CompilerInstance &ci,
        const std::string &outputFileName,
        llvm::tool_output_file *outputFile,
        const std::string &includeFileName) :
    CI(ci),
    OutputFileName(outputFileName),
    IncludeFileName(includeFileName),
    OutputFile(outputFile),
    OS(outputFile->os()),
    IndentCount(0),
    ErrorCount(0) {
  Init();
}

  // Destructor.
LuaBindingsConsumer::~LuaBindingsConsumer() {
}

  // Initialize consumer.
void LuaBindingsConsumer::Init() {
  SetIndent(0);
  LeafName = llvm::sys::path::filename(OutputFileName);
  BaseName = llvm::sys::path::stem(OutputFileName);
  PushMainDeclaration(BaseName);
  TopDeclaration().BindMode = BindStop;
  WriteFileHeader();
}

  // Shut down consumer.
void LuaBindingsConsumer::Reset() {
  // Write function bindings.
  WriteFunctionBindings();
  // Write file tail.
  WriteFileTail();
  while (DeclarationStack.size() != 0)
    PopMainDeclaration();
  if (ErrorCount != 0)
    llvm::errs() << "Lua bindings generation failed.";
}

  // Report an error.
void LuaBindingsConsumer::ReportError(const char *message) {
  Diagnostic &D = CI.getDiagnostics();
  unsigned DiagID = D.getCustomDiagID(Diagnostic::Error, message);
  D.Report(DiagID);
  ErrorCount++;
}

  // Report a formatted error message.
void LuaBindingsConsumer::ReportErrorFormatted(const char *message, ...) {
  char buf[2000];
  va_list args;
  va_start(args, message);
  vsnprintf(buf, sizeof(buf) - 1, message, args);
  ReportError(buf);
  va_end(args);
}

  // Handle a top leve declaration.  This is an entry point called by Clang.
void LuaBindingsConsumer::HandleTopLevelDecl(clang::DeclGroupRef DG) {
  // If we don't have an output file, bail.
  if (OutputFile == NULL)
    return;
  // Iterate over the declarations in the group.
  for (DeclGroupRef::iterator i = DG.begin(), e = DG.end(); i != e; ++i) {
    const Decl *D = *i;
    // If we are a named declaration...
    if (const NamedDecl *decl = dyn_cast<NamedDecl>(D)) {
      // If we shouldn't be bound, skip it.
      if (!ShouldBind(decl))
        continue;
      // Dispatch to bind function.
      BindDecl(decl);
    }
  }
}

  // This method is called when the ASTs for entire
  // translation unit have been parsed.
void LuaBindingsConsumer::HandleTranslationUnit(clang::ASTContext &Ctx) {
  Reset();
}

  // Do various generic tests to see if this declaration should be bound.
bool LuaBindingsConsumer::ShouldBind(const NamedDecl *decl) {
  if (decl->hasAttrs()) {
    specific_attr_iterator<AnnotateAttr> iter = decl->specific_attr_begin<AnnotateAttr>();
    specific_attr_iterator<AnnotateAttr> end = decl->specific_attr_end<AnnotateAttr>();
    for (; iter != end; ++ iter) {
      AnnotateAttr *attr = *iter;
      if (attr->getAnnotation().startswith("LuaBind"))
        return true;
      else if (attr->getAnnotation().startswith("LuaNoBind"))
        return false;
      else if (attr->getAnnotation().startswith("LuaStopBind")) {
        TopDeclaration().BindMode = BindStop;
        return false;
      }
      else if (attr->getAnnotation().startswith("LuaStartBind")) {
        TopDeclaration().BindMode = BindAll;
        return false;
      }
    }
  }
  if (decl->getKind() == Decl::Namespace) {
    return true;
  }
  if (TopDeclaration().BindMode == BindAll)
    return true;

  return false;
}

  // Dispatch function for binding declarations.
void LuaBindingsConsumer::BindDecl(const NamedDecl *decl) {
  Decl::Kind kind = decl->getKind();
  switch (kind) {
    case Decl::Namespace:
      {
        const NamespaceDecl *namesp = dyn_cast<NamespaceDecl>(decl);
        BindNamespace(namesp);
      }
      break;
    case Decl::Function:
      {
        const FunctionDecl *function = dyn_cast<FunctionDecl>(decl);
        AddFunction(function);
      }
      break;
    case Decl::CXXRecord:
      {
        const CXXRecordDecl *record = dyn_cast<CXXRecordDecl>(decl);
        if (record->isDefinition())
            BindClassOrStructure(record);
      }
      break;
    case Decl::CXXMethod:
      {
        const CXXMethodDecl *method = dyn_cast<CXXMethodDecl>(decl);
        if (method->isOverloadedOperator())
            break;
        AddMethod(method);
      }
      break;
    case Decl::CXXConstructor:
      {
        const CXXConstructorDecl *method = dyn_cast<CXXConstructorDecl>(decl);
        TopDeclaration().ConstructorDeclarations.push_back(method);
      }
      break;
    case Decl::CXXDestructor:
      break;
    case Decl::Record:
      {
        const RecordDecl *record = dyn_cast<RecordDecl>(decl);
        if (record->isDefinition())
            BindClassOrStructure(record);
      }
      break;
    case Decl::Enum:
      {
        const EnumDecl *enumDecl = dyn_cast<EnumDecl>(decl);
        BindEnum(enumDecl);
      }
      break;
    case Decl::Var:
      {
        const VarDecl *variable = dyn_cast<VarDecl>(decl);
        BindVariable(variable, false);
      }
      break;
    case Decl::Field:
      {
        const FieldDecl *variable = dyn_cast<FieldDecl>(decl);
        BindVariable(variable, true);
      }
      break;
    case Decl::EnumConstant:
    case Decl::ClassTemplate:
    case Decl::LinkageSpec:
    case Decl::AccessSpec:
    case Decl::Typedef:
    default:
      break;
  }
}

  // Bind a namespace.
void LuaBindingsConsumer::BindNamespace(const clang::NamespaceDecl *namesp) {
  // Push declaration onto declaration stack.
  PushNamespaceDeclaration(namesp);
  // See if we should do selective or all binding.
  if (namesp->hasAttrs()) {
    specific_attr_iterator<AnnotateAttr> aiter = namesp->specific_attr_begin<AnnotateAttr>();
    specific_attr_iterator<AnnotateAttr> aend = namesp->specific_attr_end<AnnotateAttr>();
    for (; aiter != aend; ++ aiter) {
      AnnotateAttr *attr = *aiter;
      if (attr->getAnnotation().startswith("LuaBind"))
        TopDeclaration().BindMode = BindAll;
      else
        TopDeclaration().BindMode = BindSome;
    }
  }
  else
    TopDeclaration().BindMode = BindSome;
  // Walk namespace members.
  DeclContext::decl_iterator iter = namesp->decls_begin();
  DeclContext::decl_iterator end = namesp->decls_end();
  for (; iter != end; ++iter)
  {
    // If we are a named declaration...
    if (const NamedDecl *decl = dyn_cast<NamedDecl>(*iter)) {
      // If we shouldn't be bound, skip it.
      if (!ShouldBind(decl))
        continue;
      // Dispatch to bind function.
      BindDecl(decl);
    }
  }
  // Write function bindings.
  WriteFunctionBindings();
  // Pop declaration stack.
  PopNamespaceDeclaration();
}

  // Bind a class or structure.
void LuaBindingsConsumer::BindClassOrStructure(
    const clang::RecordDecl *record) {
  OS << "\n";
  // Push declaration onto declaration stack.
  PushTypeDeclaration(record);
  // Write bind class header.
  OS << Indent << "class " << BindTypeName() << " {\n";
  // Increment indentation.
  PushIndent();
  // Write method table declaration.
  WriteMethodTableDeclaration();
  // Walk class members.
  BindRecordMembers(record, 0);
  // Write function bindings.
  WriteFunctionBindings();
  // Write class object create function.
  WriteClassCreateFunction();
  // Write get and check self function.
  WriteClassCheckFunction();
  // Write class object dispose function.
  WriteClassDisposeFunction();
  // Write public section tag.
  WriteClassPublic();
  // Write register function.
  WriteClassRegisterFunction();
    // Write class object adopt function.
  WriteClassAdoptFunction();
    // Write class object orphan function.
  WriteClassOrphanFunction();
  // Decrement indentation.
  PopIndent();
  OS << "};\n";
  // Write method table.
  WriteMethodTable();
  // Pop declaration stack.
  PopTypeDeclaration();
}

  // Bind the members of a class or structure.
void LuaBindingsConsumer::BindRecordMembers(
    const clang::RecordDecl *record, int nestLevel) {
  const clang::CXXRecordDecl *cxxRecord = dyn_cast<const CXXRecordDecl>(record);
  if (cxxRecord != NULL) {
    // Recursively walk base classes.
    CXXRecordDecl::base_class_const_iterator biter = cxxRecord->bases_begin();
    CXXRecordDecl::base_class_const_iterator bend = cxxRecord->bases_end();
    for (; biter != bend; ++biter) {
      const CXXBaseSpecifier &baseSpec(*biter);
      if (baseSpec.getAccessSpecifier() != AS_public)
        continue;
      QualType type = baseSpec.getType();
      const CXXRecordDecl *baseDecl = type->getAsCXXRecordDecl();
      if (baseDecl != NULL)
        BindRecordMembers(baseDecl, nestLevel + 1);
    }
  }
  // Walk class members.
  CXXRecordDecl::decl_iterator iter = record->decls_begin();
  CXXRecordDecl::decl_iterator end = record->decls_end();
  for (; iter != end; ++iter)
  {
    // If we are a named declaration...
    if (const NamedDecl *decl = dyn_cast<NamedDecl>(*iter)) {
      if (decl->getAccess() != AS_public)
        continue;
      // Don't do constructors.
      switch (decl->getKind()) {
        case Decl::CXXConstructor:
        case Decl::CXXDestructor:
          if (nestLevel != 0)
            continue;
          break;
        default:
          break;
      }
      // If we shouldn't be bound, skip it.
      if (!ShouldBind(decl))
        continue;
      // Dispatch to bind function.
      BindDecl(decl);
    }
  }
}

  // Bind a function or class method.
void LuaBindingsConsumer::BindFunctionOrMethod(
  FunctionOverloads *functionDesc, bool isMethod)
{
  const clang::FunctionDecl *function;
  const std::string &functionName = functionDesc->FunctionName;
  int offset = (isMethod ? 1 : 0);
  int overloadIndex;
  int overloadCount = functionDesc->Overloads.size();
  // Write bind function header.
  if (isMethod)
    OS << Indent << "static int " << functionName << "(lua_State *L) {\n";
  else {
    OS << "\n";
    OS << Indent << "int " << BaseName << "_" << functionName << "(lua_State *L) {\n";
  }
  PushIndent();
  // Get self.
  if (isMethod)
    OS << Indent << QualifiedTypeName() << " *self = Check(L, 1);\n";
  OS << Indent << "int stackCount = lua_gettop(L);\n";
  for (overloadIndex = 0; overloadIndex < overloadCount; overloadIndex++) {
    function = functionDesc->Overloads[overloadIndex];
    ParmVarDecl *paramDecl;
    QualType paramType;
    QualType functionReturnType = function->getCallResultType();
    std::string returnType = functionReturnType.getAsString();
    bool hasNonVoidReturn = !functionReturnType->isVoidType();
    bool hasRecordReturn = functionReturnType->isRecordType();
    int parameterIndex;
    int parameterCount = function->param_size();
    FunctionDecl::param_const_iterator iter;
    FunctionDecl::param_const_iterator end;
    if (overloadIndex == 0)
      OS << Indent << "if";
    else
      OS << Indent << "else if";
    PushIndent();
    if (parameterCount == 0)
      OS << " (stackCount == " << offset << ") {\n";
    else {
      OS << " ((stackCount == " << (parameterCount + offset) << ")\n";
      PushIndent();
      iter = function->param_begin();
      end = function->param_end();
      for (parameterIndex = 1; iter != end; ++iter, parameterIndex++) {
        OS << Indent << "&& ";
        WriteStackItemMatch((*iter)->getType(), parameterIndex + offset);
        if (parameterIndex == parameterCount)
          OS << ") {\n";
        else
          OS << "\n";
      }
      PopIndent();
      // Push arguments.
      iter = function->param_begin();
      end = function->param_end();
      for (parameterIndex = 1; iter != end; ++iter, parameterIndex++) {
        paramDecl = *iter;
        paramType = paramDecl->getType();
        std::string paramTypeName = paramType.getAsString();
        if (paramType->isRecordType())
          paramTypeName += " *";
        std::string paramName = paramDecl->getNameAsString();
        OS << Indent << paramTypeName << " " << paramName << " = ";
        WriteGet((*iter)->getType(), parameterIndex + offset, false);
        OS << ";\n";
      }
    }
    // Call target function.
    WriteIndent();
    if (hasNonVoidReturn) {
      if (hasRecordReturn)
        OS << returnType << " *returnValue = new "
          << StripToQualifiedName(functionReturnType) << "(";
      else
        OS << returnType << " returnValue = ";
    }
    if (isMethod)
      OS << "self->" << functionName << "(";
    else
      OS << function->getQualifiedNameAsString() << "(";
    // Push arguments.
    parameterIndex = 1;
    end = function->param_end();
    for (iter = function->param_begin(); iter != end; ++iter, ++parameterIndex) {
      paramDecl = *iter;
      paramType = paramDecl->getType();
      if (parameterIndex != 1)
        OS << ", ";
      std::string paramName = paramDecl->getNameAsString();
      if (paramType->isRecordType())
        OS << "*";
      OS << paramName;
    }
    if (hasNonVoidReturn && hasRecordReturn)
      OS << "));\n";
    else
      OS << ");\n";
    // If not void return...
    if (hasNonVoidReturn) {
      // Push return value.
      WriteIndent();
      WritePush(functionReturnType, "returnValue");
      OS << ";\n";
    }
    PopIndent();
    OS << Indent << "}\n";
  }
  OS << Indent << "else\n";
  OS << Indent << "  FunctionMatchError(L, \"" << functionName << "\");\n";
  // Return 1.
  OS << Indent << "return 1;\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Bind a variable or class or structure data member.
void LuaBindingsConsumer::BindVariable(
    const clang::DeclaratorDecl *variable, bool isField) {
  const std::string &variableName = variable->getNameAsString();
  QualType variableType = variable->getType();
  std::string variableTypeString = variableType.getAsString();
  std::string variableTypeStringStripped =
    StripToQualifiedName(variableType);
  bool isPointerType = variableType->isPointerType();
  bool isRecordType =
    (isPointerType ?
      variableType->getPointeeType()->isRecordType() :
      variableType->isRecordType());
  int offset = (isField ? 1 : 0);
  // Write bind function header.
  if (isField)
    OS << Indent << "static int " << variableName << "(lua_State *L) {\n";
  else {
    OS << "\n";
    OS << Indent << "int " << BaseName << "_" << variableName << "(lua_State *L) {\n";
  }
  PushIndent();
  // Get self.
  if (isField)
    OS << Indent << QualifiedTypeName() << " *self = Check(L, 1);\n";
  OS << Indent << "int stackCount = lua_gettop(L);\n";
  // Do getter case.
  OS << Indent << "if (stackCount == " << offset << ") { // getter\n";
  PushIndent();
  // Get value.
  if (isRecordType) {
    if (isPointerType)
      OS << Indent << variableTypeStringStripped << " *returnValue = ";
    else
      OS << Indent <<  variableTypeStringStripped << " *returnValue = new "
        << variableTypeStringStripped << "(";
  }
  else
    OS << Indent << variableTypeString << " returnValue = ";
  if (isField)
    OS << "self->" << variableName;
  else
    OS << variable->getQualifiedNameAsString();
  if (isRecordType && !isPointerType)
    OS << ");\n";
  else
    OS << ";\n";
  // Push return value.
  WriteIndent();
  WritePush(variableType, "returnValue");
  OS << ";\n";
  PopIndent();
  OS << Indent << "}\n";
  // Do setter case.
  OS << Indent << "else if ((stackCount == " << (offset + 1) << ") && ";
  WriteStackItemMatch(variableType, offset + 1);
  OS << ") // setter\n";
  PushIndent();
  // Set value.
  OS << Indent;
  if (isField)
    OS << "self->" << variableName;
  else
    OS << variable->getQualifiedNameAsString();
  OS << " = ";
  WriteGet(variableType, offset + 1, true);
  OS << ";\n";
  PopIndent();
  // Do error case.
  OS << Indent << "else\n";
  OS << Indent << "  FunctionMatchError(L, \"" << variableName << "\");\n";
  // Return 1.
  OS << Indent << "return 1;\n";
  PopIndent();
  OS << Indent << "}\n";
  // Save variable name so we can register its bind function.
  AddVariableName(variableName, isField);
}

  // Bind an enum.
void LuaBindingsConsumer::BindEnum(const clang::EnumDecl *enumDecl) {
  std::string enumName(enumDecl->getNameAsString());
  std::string functionName = BaseName + "_" + enumName;
  std::string intTypeName(enumDecl->getIntegerType().getAsString());
  OS << "\n";
  // Write bind class header.
  OS << Indent << "void " << functionName << "(lua_State *L) {\n";
  // Increment indentation.
  PushIndent();
  // Create table.
  OS << Indent << "lua_newtable(L);\n";
  // Walk enum constants.
  clang::EnumDecl::enumerator_iterator iter;
  clang::EnumDecl::enumerator_iterator end;
  for (iter = enumDecl->enumerator_begin(); iter != end; ++iter) {
    const clang::EnumConstantDecl *enumConstDecl = *iter;
    std::string enumConstName(enumConstDecl->getNameAsString());
    OS << Indent << "lua_pushliteral(L, \"" << enumConstName << "\");\n";
    OS << Indent << "lua_pushnumber(L, (" << intTypeName << ")"
      << QualifiedTypeName() << "::" << enumConstName << ");\n";
    OS << Indent << "lua_settable(L, -3);\n";
  }
  OS << Indent << "lua_setglobal(L, \"" << enumName << "\");\n";
  // Decrement indentation.
  PopIndent();
  OS << "}\n";
  AddEnumName(enumName);
}

  // Write the file header.
void LuaBindingsConsumer::WriteFileHeader() {
  // Write file header comment.
  OS << "// " << LeafName << ": Generated by the LuaBindings plugin.\n";
  OS << "// Usage: Clang -cc1 -load (path)/LuaBindings.[so|dll] -plugin luabindings \\ \n";
  OS << "//          -plugin-arg-luabindings -out=(bindingsFile).cpp \\ \n";
  OS << "//          [-plugin-arg-luabindings -include=(includeFile).cpp] \\ \n";
  OS << "//          [-x c++] (inputFile).[h|cpp]\n";
  OS << "\n";
  OS << "extern \"C\" {\n";
  OS << "#include <string.h>\n";
  OS << "#include \"lua.h\"\n";
  OS << "#include \"lauxlib.h\"\n";
  OS << "#include \"lualib.h\"\n";
  OS << "}\n";
  OS << "\n";
  // Write include file.  This file contains the declarations for the
  // symbols being bound.
  if (IncludeFileName.length())
    OS << "#include \"" << IncludeFileName << "\"\n";
  else
    OS << "#include \"" << BaseName << ".h\"\n";
  OS << "\n";
  // Put all our binding stuff in its own namespace, based on the output file name.
  WriteNamespaceBegin();
  OS << "\n";
  // Write a common function for registering a class.
  WriteClassCommonRegisterFunction();
  OS << "\n";
  // Write a common function for a function call mismatch error.
  WriteFunctionMatchErrorFunction();
}

  // Write the file tail.
void LuaBindingsConsumer::WriteFileTail() {
  if (BottomDeclaration().Functions.size()) {
    OS << "\n";
    WriteFunctionTable();
  }
  OS << "\n";
  WriteOpenFunction();
  OS << "\n";
  WriteNamespaceEnd();
}

  // Write the namespace for the output file.
void LuaBindingsConsumer::WriteNamespaceBegin() {
  OS << "namespace " << BaseName << " {\n";
}

  // Write the end of the namesape for the output file.
void LuaBindingsConsumer::WriteNamespaceEnd() {
  OS << "} // namespace " << BaseName << "\n";
}

  // Write public section tag.
void LuaBindingsConsumer::WriteClassPublic() {
  PopIndent();
  OS << Indent << "public:\n";
  PushIndent();
}

  // Write class object create function.
void LuaBindingsConsumer::WriteClassCreateFunction() {
  int parameterCount = 0;
  int parameterIndex = 1;
  int constructorCount = TopDeclaration().ConstructorDeclarations.size();
  int constructorIndex = 0;
  const CXXConstructorDecl *constructor = NULL;
  CXXConstructorDecl::param_const_iterator iter;
  CXXConstructorDecl::param_const_iterator end;
  // Write function header.
  OS << Indent << "static int Create(lua_State *L) {\n";
  PushIndent();
  if (constructorCount == 0)
    // Write call to create object instance.
    OS << Indent << QualifiedTypeName() << " *self = new "
      << QualifiedTypeName() << ";\n";
  else {
    OS << Indent << QualifiedTypeName() << " *self = NULL;\n";
    OS << Indent << "int stackCount = lua_gettop(L);\n";
    for (; constructorIndex < constructorCount; constructorIndex++) {
      constructor = TopDeclaration().ConstructorDeclarations[constructorIndex];
      parameterCount = constructor->param_size();
      if (parameterCount == 0) {
        OS << Indent << "if (stackCount == 0)\n";
        PushIndent();
        // Write call to create object instance.
        OS << Indent << "self = new "
          << QualifiedTypeName() << ";\n";
        PopIndent();
      }
      else {
        if (constructorIndex == 0)
          OS << Indent << "if";
        else
          OS << Indent << "else if";
        OS << " ((stackCount == " << parameterCount << ")\n";
        PushIndent();
        PushIndent();
        iter = constructor->param_begin();
        end = constructor->param_end();
        for (parameterIndex = 1; iter != end; ++iter, parameterIndex++) {
          OS << Indent << "&& ";
          WriteStackItemMatch((*iter)->getType(), parameterIndex);
          if (parameterIndex == parameterCount)
            OS << ") {\n";
          else
            OS << "\n";
        }
        PopIndent();
        // Push arguments.
        iter = constructor->param_begin();
        end = constructor->param_end();
        for (parameterIndex = 1; iter != end; ++iter, parameterIndex++) {
          std::string paramType = (*iter)->getType().getAsString();
          std::string paramName = (*iter)->getNameAsString();
          OS << Indent << paramType << " " << paramName << " = ";
          WriteGet((*iter)->getType(), parameterIndex, false);
          OS << ";\n";
        }
        // Write call to create object instance.
        OS << Indent << "self = new " << QualifiedTypeName() << "(";
        if (parameterCount != 0) {
          // Write arguments.
          parameterIndex = 1;
          for (iter = constructor->param_begin(); iter != end; ++iter, ++parameterIndex) {
            if (parameterIndex != 1)
              OS << ", ";
            std::string paramName = (*iter)->getNameAsString();
            OS << paramName;
          }
        }
        OS << ");\n";
        PopIndent();
        OS << Indent << "}\n";
      }
    }
    OS << Indent << "else\n";
    OS << Indent << "  FunctionMatchError(L, \"" << TypeName() << "\");\n";
  }
  // Link object instance to table.
  OS << Indent << "*(void **)lua_newuserdata(L, sizeof(void *)) = self;\n";
  OS << Indent << "luaL_getmetatable(L, \"" << TypeName() << "\");\n";
  OS << Indent << "lua_setmetatable(L, -2);\n";
  // Write return.
  OS << Indent << "return 1;\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Write class object adopt function.
void LuaBindingsConsumer::WriteClassAdoptFunction() {
  // Write function header.
  OS << Indent << "static int Adopt(lua_State *L, " << QualifiedTypeName() << " *obj) {\n";
  PushIndent();
  // Create new user data.
  OS << Indent << "*(void **)lua_newuserdata(L, sizeof(void *)) = obj;\n";
  // Associate it with a table.
  OS << Indent << "luaL_getmetatable(L, \"" << TypeName() << "\");\n";
  OS << Indent << "lua_setmetatable(L, -2);\n";
  OS << Indent << "return 1;\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Write class object orphan function.
void LuaBindingsConsumer::WriteClassOrphanFunction() {
  // Write function header.
  OS << Indent << "static int Orphan(lua_State *L) {\n";
  PushIndent();
  OS << Indent << QualifiedTypeName() << " **objp = (" << QualifiedTypeName() << " **)luaL_checkudata(L, 1, \"" << TypeName() << "\");\n";
  OS << Indent << "*objp = NULL;\n";
  OS << Indent << "return 1;\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Write class object get and check function.
void LuaBindingsConsumer::WriteClassCheckFunction() {
  // Write function header.
  OS << Indent << "static " << QualifiedTypeName() << " *Check(lua_State *L, int narg) {\n";
  PushIndent();
  // Write self type check.
  OS << Indent << "luaL_checktype(L, narg, LUA_TUSERDATA);\n";
  // Write call to get self pointer.
  OS << Indent << "void *self = luaL_checkudata(L, narg, \"" << TypeName() << "\");\n";
  OS << Indent << "if (!self)\n";
  OS << Indent << "  luaL_typerror(L, narg, \"" << TypeName() << "\");\n";
  // Write return.
  OS << Indent << "return *(" << QualifiedTypeName() << "**)self;\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Write class object dispose function.
void LuaBindingsConsumer::WriteClassDisposeFunction() {
  // Write function header.
  OS << Indent << "static int Dispose(lua_State *L) {\n";
  PushIndent();
  // Write call to get self pointer.
  OS << Indent << QualifiedTypeName() << " *self = ("
    << QualifiedTypeName() << "*)(*(void **)(lua_touserdata(L, 1)));\n";
  // Write call to delete self.
  OS << Indent << "delete self;\n";
  // Write return.
  OS << Indent << "return 0;\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Write register function.
void LuaBindingsConsumer::WriteClassRegisterFunction() {
  // Write function header.
  OS << Indent << "static int Register(lua_State *L) {\n";
  PushIndent();
  // Write call to common register function.
  OS << Indent << "return RegisterClass(L, \"" << TypeName()
    << "\", MethodTable, Dispose, Create);\n";
  PopIndent();
  OS << Indent << "}\n";
}

  // Write common class register function.
void LuaBindingsConsumer::WriteClassCommonRegisterFunction() {
  // Write function header.
  OS << Indent << "int RegisterClass(lua_State *L, const char *className," <<
      " const luaL_reg *methodTable,\n";
  OS << Indent << "    int (*disposeFunction)(lua_State *L)," <<
      " int (*createFunction)(lua_State *L)) {\n";
  // Increment indentation.
  PushIndent();
  // Write body.
  OS << Indent << "lua_newtable(L);\n";
  OS << Indent << "int methodTableIndex = lua_gettop(L);\n";
  OS << Indent << "luaL_newmetatable(L, className);\n";
  OS << Indent << "int metaTable   = lua_gettop(L);\n";
  OS << Indent << "lua_pushliteral(L, \"__metatable\");\n";
  OS << Indent << "lua_pushvalue(L, methodTableIndex);\n";
  OS << Indent << "lua_settable(L, metaTable);\n";
  OS << Indent << "lua_pushliteral(L, \"__index\");\n";
  OS << Indent << "lua_pushvalue(L, methodTableIndex);\n";
  OS << Indent << "lua_settable(L, metaTable);\n";
  OS << Indent << "lua_pushliteral(L, \"__gc\");\n";
  OS << Indent << "lua_pushcfunction(L, disposeFunction);\n";
  OS << Indent << "lua_settable(L, metaTable);\n";
  OS << Indent << "lua_pop(L, 1);\n";
  OS << Indent << "luaL_register(L, 0, methodTable);\n";
  OS << Indent << "lua_pop(L, 1);\n";
  OS << Indent << "lua_register(L, className, createFunction);\n";
  OS << Indent << "return 1;\n";
  // Decrement indentation.
  PopIndent();
  // Write function tail.
  OS << Indent << "}\n";
}

  // Write common function match error function.
void LuaBindingsConsumer::WriteFunctionMatchErrorFunction() {
  // Write function header.
  OS << Indent << "void FunctionMatchError(lua_State *L, const char *functionName) {\n";
  // Increment indentation.
  PushIndent();
  // Write body.
  OS << Indent << "char buffer[300];\n";
  OS << Indent << "strcpy(buffer, \"No matching function call signature for \");\n";
  OS << Indent << "strncat(buffer, functionName, 256);\n";
  OS << Indent << "strcat(buffer, \".\");\n";
  OS << Indent << "lua_pushstring(L, buffer);\n";
  OS << Indent << "lua_error(L);\n";
  // Decrement indentation.
  PopIndent();
  // Write function tail.
  OS << Indent << "}\n";
}

  // Write function bindings.
void LuaBindingsConsumer::WriteFunctionBindings() {
  std::vector<FunctionOverloads *>::iterator iter;
  std::vector<FunctionOverloads *>::iterator end = TopDeclaration().Functions.end();
  for (iter = TopDeclaration().Functions.begin(); iter != end; ++iter)
    BindFunctionOrMethod(*iter, (*iter)->IsMethod);
}

  // Write the method table declaration.
void LuaBindingsConsumer::WriteMethodTableDeclaration() {
  OS << Indent << "static const luaL_reg MethodTable[];\n";
}

  // Write the current table of methods to be register in Lua.
void LuaBindingsConsumer::WriteMethodTable() {
  std::vector<FunctionOverloads *>::iterator iter;
  std::vector<FunctionOverloads *>::iterator E;
  OS << "const luaL_reg " << BindTypeName() << "::MethodTable[] = {\n";
  for (
      iter = TopDeclaration().Functions.begin(),
        E = TopDeclaration().Functions.end();
      iter != E; ++iter) {
    OS << "  {\"" << (*iter)->FunctionName << "\", " << BindTypeName()
      << "::" << (*iter)->FunctionName << "},\n";
  }
  std::vector<std::string>::iterator viter;
  std::vector<std::string>::iterator vend;
  for (
      viter = TopDeclaration().VariableNames.begin(),
        vend = TopDeclaration().VariableNames.end();
      viter != vend; ++viter) {
    OS << "  {\"" << *viter << "\", " << BindTypeName()
      << "::" << *viter << "},\n";
  }
  const char *orphanName =
    ((TopDeclaration().FindFunction("Orphan") == NULL) ? "Orphan" : "_Orphan");
  OS << "  {\"" << orphanName << "\", " << BindTypeName() << "::Orphan},\n";
  OS << "  {NULL, NULL}\n";
  OS << "};\n";
}

  // Write the table of functions to be register in Lua.
void LuaBindingsConsumer::WriteFunctionTable() {
  std::vector<FunctionOverloads *>::iterator iter;
  std::vector<FunctionOverloads *>::iterator E;
  OS << "const luaL_reg FunctionTable[] = {\n";
  for (
      iter = BottomDeclaration().Functions.begin(),
        E = BottomDeclaration().Functions.end();
      iter != E; ++iter) {
    OS << "  {\"" << (*iter)->FunctionName << "\", " << BindTypeName()
      << "_" << (*iter)->FunctionName << "},\n";
  }
  std::vector<std::string>::iterator viter;
  std::vector<std::string>::iterator vend;
  for (
      viter = TopDeclaration().VariableNames.begin(),
        vend = TopDeclaration().VariableNames.end();
      viter != vend; ++viter) {
    OS << "  {\"" << *viter << "\", " << BindTypeName()
      << "_" << *viter << "},\n";
  }
  OS << "  {NULL, NULL}\n";
  OS << "};\n";
}

  // Write the open function for registering everything in the output file.
void LuaBindingsConsumer::WriteOpenFunction() {
  std::vector<std::string>::iterator iter;
  std::vector<std::string>::iterator E;
  OS << "void Open(lua_State *L) {\n";
  for (
      iter = TopDeclaration().EnumNames.begin(),
      E = TopDeclaration().EnumNames.end();
      iter != E; ++iter) {
    OS << "  " << BaseName << "_" << *iter << "(L);\n";
  }
  for (
      iter = TopDeclaration().StructureNames.begin(),
        E = TopDeclaration().StructureNames.end();
      iter != E; ++iter) {
    OS << "  " << BaseName << "_" << *iter << "::Register(L);\n";
  }
  for (
      iter = TopDeclaration().TypeNames.begin(),
      E = TopDeclaration().TypeNames.end();
      iter != E; ++iter) {
    OS << "  " << BaseName << "_" << *iter << "::Register(L);\n";
  }
  if (BottomDeclaration().Functions.size()) {
    OS << "  " << "luaL_register(L, \"" << BaseName << "\", FunctionTable);\n";
  }
  OS << "}\n";
}

  // Write a stack item match expression.
void LuaBindingsConsumer::WriteStackItemMatch(QualType type, int stackIndex) {
  if (type->isPointerType()) {
    QualType pointeeType = type->getPointeeType();
    if (pointeeType->isCharType()) {
      OS << "lua_isstring(L, " << stackIndex << ")";
    }
    else if (pointeeType->isRecordType()) {
      OS << "lua_isuserdata(L, " << stackIndex << ")";
    }
    else {
      // Treat other pointers a integers, passing the address.
      OS << "lua_isnumber(L, " << stackIndex << ")";
    }
  }
  else if (type->isRecordType()) {
    OS << "luaL_checkudata(L, " << stackIndex << ", \""
      << StripToBaseName(type) << "\")";
  }
  else if (type->isBooleanType()) {
    OS << "lua_isboolean(L, " << stackIndex << ")";
  }
  else if (type->isEnumeralType()) {
    OS << "lua_isnumber(L, " << stackIndex << ")";
  }
  else if (type->isIntegerType()) {
    OS << "lua_isnumber(L, " << stackIndex << ")";
  }
  else if (type->isFloatingType()) {
    OS << "lua_isnumber(L, " << stackIndex << ")";
  }
  else {
    ReportErrorFormatted("Can't map type %s.", type.getAsString().c_str());
  }
}

  // Write the function call for getting an argument from the Lua stack.
void LuaBindingsConsumer::WriteGet(QualType type, int stackIndex, bool forVariableAssign) {
  if (type->isPointerType()) {
    QualType pointeeType = type->getPointeeType();
    if (pointeeType->isCharType()) {
      OS << "(" << type.getAsString() << ")luaL_checkstring(L, " << stackIndex << ")";
    }
    else if (pointeeType->isRecordType()) {
      if (forVariableAssign)
        OS << "*(" << StripToQualifiedName(type) << "**)" << "luaL_checkudata(L, " << stackIndex
          << ", \"" << StripToBaseName(type) << "\")";
      else
        OS << "*(" << StripToQualifiedName(type) << "**)" << "luaL_checkudata(L, " << stackIndex
          << ", \"" << StripToBaseName(type) << "\")";
    }
    else {
      // Treat other pointers a integers, passing the address.
      OS << "(" << type.getAsString() << ")luaL_checkinteger(L, " << stackIndex << ")";
    }
  }
  else if (type->isRecordType()) {
    if (forVariableAssign)
      OS << "**(" << StripToQualifiedName(type) << " **)" << "luaL_checkudata(L, " << stackIndex
        << ", \"" << StripToBaseName(type) << "\")";
    else
      OS << "*(" << StripToQualifiedName(type) << " **)" << "luaL_checkudata(L, " << stackIndex
        << ", \"" << StripToBaseName(type) << "\")";
  }
  else if (type->isBooleanType()) {
    OS << "(luaL_checkboolean(L, " << stackIndex << ") ? true : false)";
  }
  else if (type->isEnumeralType()) {
    OS << "(" << type.getAsString() << ")luaL_checkinteger(L, " << stackIndex << ")";
  }
  else if (type->isIntegerType()) {
    OS << "(" << type.getAsString() << ")luaL_checkinteger(L, " << stackIndex << ")";
  }
  else if (type->isFloatingType()) {
    OS << "(" << type.getAsString() << ")luaL_checknumber(L, " << stackIndex << ")";
  }
  else {
    ReportErrorFormatted("Can't map type %s.", type.getAsString().c_str());
  }
}

  // Write the function call for pushing a value on the Lua stack.
void LuaBindingsConsumer::WritePush(QualType type, const char *variableName) {
  if (type->isPointerType()) {
    QualType pointeeType = type->getPointeeType();
    if (pointeeType->isCharType()) {
      OS << "lua_pushstring(L, " << variableName << ")";
    }
    else if (pointeeType->isRecordType()) {
      OS << GetBindClassNameFromType(pointeeType) << "::Adopt(L, " << variableName << ")";
    }
    else {
      OS << "lua_pushinteger(L, (size_t)" << variableName << ")";
    }
  }
  else if (type->isRecordType()) {
    OS << GetBindClassNameFromType(type) << "::Adopt(L, " << variableName << ")";
  }
  else if (type->isBooleanType()) {
    OS << "lua_pushboolean(L, (" << variableName << " ? 1 : 0))";
  }
  else if (type->isEnumeralType()) {
    OS << "lua_pushinteger(L, (lua_Integer)" << variableName << ")";
  }
  else if (type->isIntegerType()) {
    OS << "lua_pushinteger(L, " << variableName << ")";
  }
  else if (type->isFloatingType()) {
    OS << "lua_pushnumber(L, " << variableName << ")";
  }
  else
    ReportErrorFormatted("Can't map type %s.", type.getAsString().c_str());
}

  // Write a line indent.
void LuaBindingsConsumer::WriteIndent() {
  OS << Indent;
}

  // Set the line indent.
void LuaBindingsConsumer::SetIndent(int indentCount) {
  const int IndentSize = 2;
  IndentCount = indentCount;
  int charCount = IndentCount * IndentSize;
  if (charCount > (int)sizeof(Indent) - 2) {
    charCount = sizeof(Indent) - 2;
    IndentCount = charCount/IndentSize;
  }
  else if (charCount < 0)
    charCount = IndentCount = 0;
  memset(Indent, ' ', charCount);
  Indent[charCount] = '\0';
}

  // Offset the line indent.
void LuaBindingsConsumer::OffsetIndent(int offset) {
  SetIndent(IndentCount + offset);
}

  // Increment the line indent.
void LuaBindingsConsumer::PushIndent() {
  OffsetIndent(1);
}

  // Decrement the line indent.
void LuaBindingsConsumer::PopIndent() {
  OffsetIndent(-1);
}

  // Strip a type of everything but the base type.
clang::QualType LuaBindingsConsumer::StripToBaseType(clang::QualType type) {
  for (;;) {
    if (type->isPointerType())
      type = type->getPointeeType();
    else if (type->isReferenceType())
      type = type.getNonReferenceType();
    // What else can I strip?
    else
      break;
  }
  type = type.getUnqualifiedType();
  return type;
}

  // Strip a type of everything but the base type name.
std::string LuaBindingsConsumer::StripToBaseName(clang::QualType type) {
  std::string typeName = StripToQualifiedName(type);
  size_t index;
  while ((index = typeName.find_first_of("::")) < typeName.length())
    typeName = typeName.substr(index + 2);
  return typeName;
}

  // Strip a type of everything but the qualified base type name.
std::string LuaBindingsConsumer::StripToQualifiedName(clang::QualType type) {
  size_t index;
  type = StripToBaseType(type);
  std::string typeName = type.getAsString();
  if (typeName.compare(0, 6, "class ") == 0)
    typeName = typeName.substr(6);
  else if (typeName.compare(0, 7, "struct ") == 0)
    typeName = typeName.substr(7);
  else if (typeName.compare(0, 5, "enum ") == 0)
    typeName = typeName.substr(5);
  if ((index = typeName.find("::struct ")) < typeName.length())
    typeName = typeName.erase(index + 2, 7);
  if ((index = typeName.find("::class ")) < typeName.length())
    typeName = typeName.erase(index + 2, 6);
  if ((index = typeName.find("::enum ")) < typeName.length())
    typeName = typeName.erase(index + 2, 5);
  while ((index = typeName.find_first_of('[')) < typeName.length())
    typeName = typeName.substr(0, index);
  while ((index = typeName.find_first_of('{')) < typeName.length())
    typeName = typeName.substr(0, index);
  while ((index = typeName.find_first_of(' ')) < typeName.length())
    typeName = typeName.substr(0, index);
  while ((index = typeName.find_first_of('\t')) < typeName.length())
    typeName = typeName.substr(0, index);
  return typeName;
}

  // Get the bind class name from a type.
std::string LuaBindingsConsumer::GetBindClassNameFromType(clang::QualType type) {
  return std::string(BaseName + "_" + StripToBaseName(type));
}

  // Add a variable name to be registered.
void LuaBindingsConsumer::AddVariableName(const std::string &name, bool isField) {
  if (isField)
    TopDeclaration().VariableNames.push_back(name);
  else
    BottomDeclaration().VariableNames.push_back(name);
}

  // Add a function to be registered.
void LuaBindingsConsumer::AddFunction(const clang::FunctionDecl *functionDecl) {
  FunctionOverloads *functionOverloads = BottomDeclaration().FindFunction(
    functionDecl->getNameAsString());
  if (functionOverloads != NULL)
    functionOverloads->Overloads.push_back(functionDecl);
  else
    BottomDeclaration().Functions.push_back(
      new FunctionOverloads(
        functionDecl, functionDecl->getNameAsString(), false));
}

  // Add a method to be registered.
void LuaBindingsConsumer::AddMethod(const clang::CXXMethodDecl *methodDecl) {
  FunctionOverloads *functionOverloads = TopDeclaration().FindFunction(
    methodDecl->getNameAsString());
  if (functionOverloads != NULL)
    functionOverloads->Overloads.push_back(methodDecl);
  else
    TopDeclaration().Functions.push_back(
      new FunctionOverloads(
        methodDecl, methodDecl->getNameAsString(), true));
}

  // Add an enum register function name to be registered.
void LuaBindingsConsumer::AddEnumName(const std::string &name) {
  BottomDeclaration().EnumNames.push_back(name);
}

  // Add a type name to be registered.
void LuaBindingsConsumer::AddTypeName(const std::string &name) {
  BottomDeclaration().TypeNames.push_back(name);
}

  // Push a type declaration on the declaration stack.
void LuaBindingsConsumer::PushTypeDeclaration(const clang::NamedDecl *decl) {
  std::string typeName(decl->getNameAsString());
  std::string qualifiedTypeName(decl->getQualifiedNameAsString());
  std::string bindTypeName(BaseName + "_" + typeName);
  DeclarationState *declaration = new DeclarationState(decl, typeName, qualifiedTypeName, bindTypeName);
  DeclarationStack.push_back(declaration);
}

  // Pop a type declaration from the declaration stack.
void LuaBindingsConsumer::PopTypeDeclaration() {
  assert(DeclarationStack.size() != 0);
  DeclarationState * declarationState = GetDeclaration(0);
  DeclarationStack.pop_back();
  AddTypeName(declarationState->TypeName);
  delete declarationState;
}

  // Push a namespace declaration on the declaration stack.
void LuaBindingsConsumer::PushNamespaceDeclaration(const clang::NamedDecl *decl) {
  std::string typeName(decl->getNameAsString());
  std::string qualifiedTypeName(decl->getQualifiedNameAsString());
  DeclarationState *declaration = new DeclarationState(decl, typeName, qualifiedTypeName, typeName);
  DeclarationStack.push_back(declaration);
}

  // Pop a namespace declaration from the declaration stack.
void LuaBindingsConsumer::PopNamespaceDeclaration() {
  assert(DeclarationStack.size() != 0);
  DeclarationState * declarationState = GetDeclaration(0);
  DeclarationStack.pop_back();
  delete declarationState;
}

  // Push a declaration on the declaration stack.
void LuaBindingsConsumer::PushMainDeclaration(const std::string &baseName) {
  DeclarationState *declaration = new DeclarationState(baseName);
  DeclarationStack.push_back(declaration);
}

  // Pop a declaration from the declaration stack.
void LuaBindingsConsumer::PopMainDeclaration() {
  assert(DeclarationStack.size() != 0);
  DeclarationState * declarationState = GetDeclaration(0);
  DeclarationStack.pop_back();
  delete declarationState;
}

  // Get a declaration from the declaration stack relative to the stack top.
  // 0 is top of stack.
DeclarationState *LuaBindingsConsumer::GetDeclaration(int index) {
  index = DeclarationStack.size() - (index + 1);
  if ((index >= 0) && (index < (int)DeclarationStack.size()))
    return DeclarationStack[index];
  return NULL;
}

  // Get the top declaration from the declaration stack.
DeclarationState &LuaBindingsConsumer::TopDeclaration() {
  assert(DeclarationStack.size() != 0);
  return *GetDeclaration(0);
}

  // Get the bottom declaration from the declaration stack.
DeclarationState &LuaBindingsConsumer::BottomDeclaration() {
  assert(DeclarationStack.size() != 0);
  return *DeclarationStack[0];
}

} // namespace LuaBindings
