//===- LuaBindingsConsumer.h ---------------------------------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// LuaBindingConsumer.h - Declaration of LuaBindingsConsumer.
//
//===----------------------------------------------------------------------===//

namespace LuaBindings {

class FunctionOverloads
{
public:
  FunctionOverloads(
    const clang::FunctionDecl *function,
    std::string functionName,
    bool isMethod);
  ~FunctionOverloads();
  std::string FunctionName;
  std::vector<const clang::FunctionDecl *> Overloads;
  bool IsMethod;
};

enum BindingMode {
  BindAll,
  BindSome,
  BindStop
};

class DeclarationState
{
public:
  DeclarationState(const
    clang::NamedDecl *decl,
    std::string &typeName,
    std::string &qualifiedTypeName,
    std::string &bindTypeName);
  DeclarationState(std::string bindTypeName);
  ~DeclarationState();
    // The associated declaration.
  const clang::NamedDecl *Declaration;
    // The current client type name.
  std::string TypeName;
    // The current client qualified type name.
  std::string QualifiedTypeName;
    // The current qualifiers.
  std::string Qualifiers;
    // The current bind type name.
  std::string BindTypeName;
    // The list of bound type names.
  std::vector<std::string> TypeNames;
    // The list of bound structure names.
  std::vector<std::string> StructureNames;
    // The list of bound variable names.
  std::vector<std::string> VariableNames;
    // The list of bound enum names.
  std::vector<std::string> EnumNames;
    // The list of function overloads.
  std::vector<FunctionOverloads *> Functions;
    // Find a function overload entry.
  FunctionOverloads *FindFunction(const std::string &name);
    // The list of constructors.
  std::vector<const clang::CXXConstructorDecl *> ConstructorDeclarations;
    // Indicates binding has been stopped.
  BindingMode BindMode;
};

class LuaBindingsConsumer : public clang::ASTConsumer {
protected:
  clang::CompilerInstance &CI;
    // Output file path given to the plugin.
  std::string OutputFileName;
    // Include file to be #include'ed in the output file.
  std::string IncludeFileName;
    // The output stream manager.
  llvm::tool_output_file *OutputFile;
    // The output stream.
  llvm::raw_fd_ostream &OS;
    // The output path file leaf name.
  std::string LeafName;
    // The output file base name (without extension).
  std::string BaseName;
    // The declaration stack.
  std::vector<DeclarationState *> DeclarationStack;
    // Current indentation count.
  int IndentCount;
    // Current indentation string.
  char Indent[256];
    // How many errors happened.
  int ErrorCount;
public:
    // Constructor.
  LuaBindingsConsumer(
    clang::CompilerInstance &ci,
    const std::string &outputFileName,
    llvm::tool_output_file *outputFile,
    const std::string &includeFileName);
    // Destructor.
  virtual ~LuaBindingsConsumer();
    // Initialize consumer.
  void Init();
    // Shut down consumer.
  void Reset();
    // Report an error.
  void ReportError(const char *message);
    // Report a formatted error message.
  void ReportErrorFormatted(const char *message, ...);
    // Handle a top leve delcaration.  This is an entry point called by Clang.
  virtual void HandleTopLevelDecl(clang::DeclGroupRef DG);
    // This method is called when the ASTs for entire
    // translation unit have been parsed.
  virtual void HandleTranslationUnit(clang::ASTContext &Ctx);
    // Do various generic tests to see if this declaration should be bound.
  bool ShouldBind(const clang::NamedDecl *decl);
    // Dispatch function for binding declarations.
  void BindDecl(const clang::NamedDecl *decl);
    // Bind a namespace.
  void BindNamespace(const clang::NamespaceDecl *namesp);
    // Bind a class or structure.
  void BindClassOrStructure(const clang::RecordDecl *record);
    // Bind the members of a class or structure.
  void BindRecordMembers(const clang::RecordDecl *record, int nestLevel);
    // Bind a function or class method.
  void BindFunctionOrMethod(
      FunctionOverloads *functionDesc, bool isMethod);
    // Bind a variable or class or structure data member.
  void BindVariable(const clang::DeclaratorDecl *variable, bool isField);
    // Bind an enum.
  void BindEnum(const clang::EnumDecl *enumDecl);
    // Write the file header.
  void WriteFileHeader();
    // Write the file tail.
  void WriteFileTail();
    // Write the namespace for the output file.
  void WriteNamespaceBegin();
    // Write the end of the namesape for the output file.
  void WriteNamespaceEnd();
    // Write public section tag.
  void WriteClassPublic();
    // Write class object create function.
  void WriteClassCreateFunction();
    // Write class object adopt function.
  void WriteClassAdoptFunction();
    // Write class object orphan function.
  void WriteClassOrphanFunction();
    // Write class object get and check function.
  void WriteClassCheckFunction();
    // Write class object dispose function.
  void WriteClassDisposeFunction();
    // Write class register function.
  void WriteClassRegisterFunction();
    // Write common class register function.
  void WriteClassCommonRegisterFunction();
    // Write common function match error function.
  void WriteFunctionMatchErrorFunction();
    // Write function bindings.
  void WriteFunctionBindings();
    // Write the method table declaration.
  void WriteMethodTableDeclaration();
    // Write the current table of methods to be register in Lua.
  void WriteMethodTable();
    // Write the table of functions to be register in Lua.
  void WriteFunctionTable();
    // Write the open function for registering everything in the output file.
  void WriteOpenFunction();
    // Write a stack item match expression.
  void WriteStackItemMatch(clang::QualType type, int stackIndex);
    // Write the function call for getting an argument from the Lua stack.
  void WriteGet(clang::QualType type, int stackIndex, bool forVariableAssign);
    // Write the function call for pushing a value on the Lua stack.
  void WritePush(clang::QualType type, const char *variableName);
    // Write a line indent.
  void WriteIndent();
    // Set the line indent.
  void SetIndent(int indentCount);
    // Offset the line indent.
  void OffsetIndent(int offset);
    // Increment the line indent.
  void PushIndent();
    // Decrement the line indent.
  void PopIndent();
    // Strip a type of everything but the base type.
  clang::QualType StripToBaseType(clang::QualType type);
    // Strip a type of everything but the base type name and return its string.
  std::string StripToBaseName(clang::QualType type);
    // Strip a type of everything but the base qualified type name and return its string.
  std::string StripToQualifiedName(clang::QualType type);
    // Get the bind class name from a type.
  std::string GetBindClassNameFromType(clang::QualType type);
    // Add a variable name to be registered.
  void AddVariableName(const std::string &name, bool isField);
    // Add a function to be registered.
  void AddFunction(const clang::FunctionDecl *functionDecl);
    // Add a method to be registered.
  void AddMethod(const clang::CXXMethodDecl *methodDecl);
    // Add an enum name to be registered.
  void AddEnumName(const std::string &name);
    // Add a type name to be registered.
  void AddTypeName(const std::string &name);
    // Push a type declaration on the declaration stack.
  void PushTypeDeclaration(const clang::NamedDecl *decl);
    // Pop a type declaration from the declaration stack.
  void PopTypeDeclaration();
    // Push a namespace declaration on the declaration stack.
  void PushNamespaceDeclaration(const clang::NamedDecl *decl);
    // Pop a namespace declaration from the declaration stack.
  void PopNamespaceDeclaration();
    // Push a declaration on the declaration stack.
  void PushMainDeclaration(const std::string &baseName);
    // Pop a declaration from the declaration stack.
  void PopMainDeclaration();
    // Get a declaration from the declaration stack relative to the stack top.
    // 0 is top of stack.
  DeclarationState *GetDeclaration(int index);
    // Get the top declaration from the declaration stack.
  DeclarationState &TopDeclaration();
    // Get the bottom declaration from the declaration stack.
  DeclarationState &BottomDeclaration();
    // Get the current bind type name.
  const std::string &BindTypeName() { return TopDeclaration().BindTypeName; }
    // Get the current client type name.
  const std::string &TypeName() { return TopDeclaration().TypeName; }
    // Get the current client qualified type name.
  const std::string &QualifiedTypeName() { return TopDeclaration().QualifiedTypeName; }
    // Get the current client type name.
  const std::string &Qualifiers() { return TopDeclaration().Qualifiers; }
    // Get the current type declaration.
  const clang::NamedDecl *Declaration() { return TopDeclaration().Declaration; }
};

} // namespace LuaBindings
