/*
** $Id: ldump.c,v 2.8.1.1 2007/12/27 13:02:25 roberto Exp $
** save precompiled Lua chunks
** See Copyright Notice in lua.h
*/

#include <stddef.h>
#include <assert.h>
#include <stdlib.h>

#define ldump_c
#define LUA_CORE

#include "lua.h"

#include "lobject.h"
#include "lstate.h"
#include "lundump.h"
#include <string.h>

typedef struct {
 lua_State* L;
 lua_Writer writer;
 void* data;
 int strip;
 int status;
 LuaDumpConfig *config;
} DumpState;

#define DumpMem(b,n,size,D)	DumpBlock(b,(n)*(size),D)

static void DumpBlock(const void* b, size_t size, DumpState* D)
{
 if (D->status==0)
 {
  lua_unlock(D->L);
  D->status=(*D->writer)(D->L,b,size,D->data);
  lua_lock(D->L);
 }
}

// Use this version on arrays
static void MemoryEndianSwap32Bit(void* pDst, const void* pSrc, size_t size)
{
	size_t index;
	size_t endianIndex;
	const unsigned char* pSrcU8 = (const unsigned char*)pSrc;
	unsigned char* pDstU8 = (unsigned char*)pDst;

	lua_assert((size % 4) == 0);

	for (index = 0; index < size; index += 4)
	{
		for (endianIndex = 0; endianIndex < 4; ++endianIndex)
		{
			pDstU8[index + endianIndex] = pSrcU8[index + 4 - 1 - endianIndex];
		}
	}
}

// Use this version on individual types
static void EndianSwap(unsigned char* pBuffer, size_t iLength)
{
	size_t i = 0;
	size_t j = iLength - 1;
	while (i < j)
	{
		unsigned char temp = pBuffer[i];
		pBuffer[i] = pBuffer[j];
		pBuffer[j] = temp;

		i++, j--;
	}
}

static void DumpChar(int y, DumpState* D)
{
 char x=(char)y;
 DumpMem(&x, 1, sizeof(x), D);
}

static void DumpInt(int x, DumpState* D)
{
	const size_t sizeBuf = D->config->sizeof_int;
	const size_t sizeX = sizeof(x);
	const size_t sizeCopy = ((sizeBuf >= sizeX) ? sizeX : sizeBuf);
	void* pBuffer = 0;

	assert((sizeBuf % 2) == 0);
	assert((sizeX % 2) == 0);

	pBuffer = malloc(sizeBuf);
	memset(pBuffer, 0x0, sizeBuf);
	memcpy(pBuffer, &x, sizeCopy);

	if (D->config->endianness != 1)
		EndianSwap((unsigned char*)pBuffer, sizeBuf);

	DumpMem(pBuffer, 1, sizeBuf, D);
	free(pBuffer);

	// Original Lua implementation:
	// DumpVar(x, D);
}

static void DumpNumber(lua_Number x, DumpState* D)
{
	// NOTE: Only really support lua_Number being a float or double

	const size_t sizeBuf = D->config->sizeof_lua_Number;
	const size_t sizeX = sizeof(x);
	const size_t sizeCopy = ((sizeBuf >= sizeX) ? sizeX : sizeBuf);
	void* pBuffer = 0;

	assert((sizeBuf % 4) == 0);
	assert((sizeX % 4) == 0);
	assert((sizeBuf >= 4) && (sizeBuf <= 8));
	assert((sizeX >= 4) && (sizeBuf <= 8));

	pBuffer = malloc(sizeBuf);
	memset(pBuffer, 0x0, sizeBuf);

	if (sizeX == sizeBuf)
		memcpy(pBuffer, &x, sizeCopy);
	else
	{
		if (sizeX == 4)
		{
			double dValue = (double)x;
			memcpy(pBuffer, &dValue, sizeCopy);
		}
		else
		{
			float flValue = (float)x;
			memcpy(pBuffer, &flValue, sizeCopy);
		}
	}

	if (D->config->endianness != 1)
		EndianSwap((unsigned char*)pBuffer, sizeBuf);

	DumpMem(pBuffer, 1, sizeBuf, D);
	free(pBuffer);

	// Original Lua implmenetation:
	// DumpVar(x, D);
}

static void DumpVector(const void* b, int n, size_t size, DumpState* D)
{
	if (D->config->endianness != 1)
	{
		void* pBuffer = malloc(n * size);
		MemoryEndianSwap32Bit(pBuffer, b, n * size);
		DumpInt(n, D);
		DumpMem(pBuffer, n, size, D);
		free(pBuffer);
	}
	else
	{
		DumpInt(n, D);
		DumpMem(b, n, size, D);
	}

	// Original Lua implementation
	// DumpInt(n, D);
	// DumpMem(b, n, size, D);
}

static void DumpString(const TString* s, DumpState* D)
{
 if (s==NULL || getstr(s)==NULL)
 {
  size_t size=0;
  //DumpVar(size,D);
  DumpInt((int)size, D);
 }
 else
 {
  size_t size=s->tsv.len+1;		/* include trailing '\0' */
  //DumpVar(size,D);
  DumpInt((int)size, D);
  DumpBlock(getstr(s),size,D);
 }
}

#define DumpCode(f,D)	 DumpVector(f->code,f->sizecode,sizeof(Instruction),D)

static void DumpFunction(const Proto* f, const TString* p, DumpState* D);

static void DumpConstants(const Proto* f, DumpState* D)
{
 int i,n=f->sizek;
 DumpInt(n,D);
 for (i=0; i<n; i++)
 {
  const TValue* o=&f->k[i];
  DumpChar(ttype(o),D);
  switch (ttype(o))
  {
   case LUA_TNIL:
	break;
   case LUA_TBOOLEAN:
	DumpChar(bvalue(o),D);
	break;
   case LUA_TNUMBER:
	DumpNumber(nvalue(o),D);
	break;
   case LUA_TSTRING:
	DumpString(rawtsvalue(o),D);
	break;
   default:
	lua_assert(0);			/* cannot happen */
	break;
  }
 }
 n=f->sizep;
 DumpInt(n,D);
 for (i=0; i<n; i++) DumpFunction(f->p[i],f->source,D);
}

static void DumpDebug(const Proto* f, DumpState* D)
{
 int i,n;
 n= (D->strip) ? 0 : f->sizelineinfo;
 DumpVector(f->lineinfo,n,sizeof(int),D);
 n= (D->strip) ? 0 : f->sizelocvars;
 DumpInt(n,D);
 for (i=0; i<n; i++)
 {
  DumpString(f->locvars[i].varname,D);
  DumpInt(f->locvars[i].startpc,D);
  DumpInt(f->locvars[i].endpc,D);
 }
 n= (D->strip) ? 0 : f->sizeupvalues;
 DumpInt(n,D);
 for (i=0; i<n; i++) DumpString(f->upvalues[i],D);
}

static void DumpFunction(const Proto* f, const TString* p, DumpState* D)
{
 DumpString((f->source==p || D->strip) ? NULL : f->source,D);
 DumpInt(f->linedefined,D);
 DumpInt(f->lastlinedefined,D);
 DumpChar(f->nups,D);
 DumpChar(f->numparams,D);
 DumpChar(f->is_vararg,D);
 DumpChar(f->maxstacksize,D);
 DumpCode(f,D);
 DumpConstants(f,D);
 DumpDebug(f,D);
}

static void DumpHeader(DumpState* D)
{
 char h[LUAC_HEADERSIZE];
 luaU_header(h, D->config);
 DumpBlock(h,LUAC_HEADERSIZE,D);
}

/*
** dump Lua function as precompiled chunk
*/
int luaU_dump (lua_State* L, const Proto* f, lua_Writer w, void* data, int strip)
{
 LuaDumpConfig config;
 DumpState D;
 D.L=L;
 D.writer=w;
 D.data=data;
 D.strip=strip;
 D.status=0;

 // Dump functions are expecting this config 
 config.endianness = 1; // guess
 config.sizeof_int = sizeof(int);
 config.sizeof_size_t = sizeof(size_t);
 config.sizeof_lua_Number = sizeof(lua_Number);
 D.config = &config;

 DumpHeader(&D);
 DumpFunction(f,NULL,&D);
 return D.status;
}

/*
** dump Lua function as precompiled chunk
*/

int luaU_dumpEx (lua_State *L, const Proto *f, lua_Writer w, void *data, int strip, LuaDumpConfig *config)
{
	DumpState D;
	D.L = L;
	D.writer = w;
	D.data = data;
	D.strip = strip;
	D.status = 0;
	D.config = config;
	DumpHeader(&D);
	DumpFunction(f, NULL, &D);

	return D.status;
}
